/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#include"tabusearch.h"
#include"spheres.h"
#include<iostream>
#include<iomanip>
#include<fstream>
#include<cstdlib>
#include<climits>
#include<cassert>
#include<cstring>
#include<cmath>
#include<algorithm>
using namespace std;
/*-------------------------------------------------------------------------------------------------+
|                                                                                                  |
|                                   local search for k-cluster                                     |
|                                                                                                  |
+-------------------------------------------------------------------------------------------------*/

/*----------------------     Project-wide Global Variables      ---------------------*/
int**   edge;       //These global variables simply define the graph:
int     n;          //vtx no
int     m;          //edge no
int     k;          //cluster size
double**ww;         //weight of edge[i][j]: 1 if {i,j}\in V, 0 otherwise
int objMaxPossible; //maximum number of edges in a k-clique

/*----------------------   Global Variables Used in tabuSearch  ---------------------*/
int* currSol;
int* tabu = NULL;
double* contrib; //current contribution of each vertex to objective, considering the
                 //currently selected vertices: contrib[i]=\sum_{j\in currSol}w[i][j]

void allocGlobalsTabuSearch();
void freeGlobalsTabuSearch();

//Linux kernel coding rules: macros resembling functions may be named in lower case
//This macro can update the choice of the best move: selected 'in' needs deselection 
//non-selected 'out' gets selected.
#define updateBestMove(i, j)                                                      \
    do{                                                                           \
        double contribNow = contrib[j]-contrib[i]-ww[i][j];                       \
        if(contribNow>bestContrib) {                                              \
             bestContrib = contribNow;                                            \
             in     = i;                                                          \
             out    = j;                                                          \
             bstRnd = rand();/*for equal quality moves, take one with max rnd*/   \
        }else{                                                                    \
            if(contribNow==bestContrib){                                          \
                int rndNow =  rand();                                             \
                if(rndNow>bstRnd) {                                               \
                     bestContrib = contribNow;                                    \
                     in     = i;                                                  \
                     out    = j;                                                  \
                     bstRnd = rndNow;                                             \
                }                                                                 \
            }                                                                     \
        }                                                                         \
    }while(false);

#define TOP_SZ 3                                      //size of the top of best diff obj vals
double platObj=0, quasiPlatHeighDiff=0;               
int movesOnQuasiPlateau(int iter, double obj){
    static double topObj[TOP_SZ];                     //the top of best diff obj vals
    static int movesPlat;
    //Reset data at iter 1000, see below function returns 0 when iter<1000
    if(iter==1) {                                  
        for(int i=0;i<TOP_SZ;i++)
            topObj[i] = 0;
        quasiPlatHeighDiff = 0;
        return 0;
    }
    //clog<<quasiPlatHeighDiff<<",";
    if(abs(platObj-obj)<quasiPlatHeighDiff)           //Calculate the number of moves
        movesPlat++;                                  //movesPlat on a quasi plateau 
    else {
        movesPlat = 0;
        platObj = obj;
    }
    //Update quasiPlatHeighDiff and best TOP_SZ objective values
    for(int i=0;i<TOP_SZ;i++){
        if(topObj[i]==obj)
            break;
        if(topObj[i]<obj){
            for(int j=TOP_SZ-1;j>i;j--)             //shift best obj values ever seen
                topObj[j] = topObj[j-1];
            topObj[i] = obj;
            if(topObj[TOP_SZ-1]!=0)                 //update plat height when new obj in top 
                quasiPlatHeighDiff = topObj[0] - topObj[TOP_SZ-1];
            break;
        }
    }
    if(iter<1000)                                     //technique disabled before iter 1000
        return 0;
    return movesPlat;
}

//Runs a local search to find a maximum weight cluster of size k (k vertices)
//Start from center, write the best found cluster in solbest, write iter(Used)
//Consider edge weights ww, some of them may be very negative (missing edge)
//run until reaching maxIter or a distance maxRadius from center
double tabuSearchSegment(int* solbest, int &iter, int* center, int maxIter, int maxRadius)
{
    //init date regarding the search (e.g., make tabu lists at 0, calculate contributions at iter 0 
    double startObj;
    double   bestObj= INT_MIN;
    double   obj    = 0;
    int currDist    = 0;  //current dist to center, used only if maxRadius!=INFTY_RADIUS_NO_DISTS
    int bestDist    = 0;  //distance from best found to center is 0 at first
    double repulsiveForce = 0;
    for(int i=0;i<n;i++) {
        tabu[i]   = 0;
        currSol[i] = 0;
        contrib[i]= 0;
    }

    for(int i=0;i<(::n);i++)
        currSol[i] = center[i];
    
    //contrib[j] is the variation of the obj when vtx j is selected
    for(int i=0;i<(::n);i++)
        if(currSol[i]){
            for(int j=i+1;j<(::n);j++)
                if(currSol[j])
                   obj+=::ww[i][j];
            for(int j=0;j<(::n);j++)
                contrib[j] +=::ww[i][j];
        }
    bestObj  = obj; 
    startObj = obj;
    //second condition is always false if maxRadius = INFTY_RADIUS_NO_DISTS
    for(iter=1; (iter<=maxIter) && (currDist<maxRadius) && (obj<objMaxPossible) ;iter++) {
        //log information from time to time  
        if( (iter%5000>=4950) || (iter%5000<=50)) {
            clog<<"Ls iter="<<iter<<" obj="<<obj<<"bestobj="<<bestObj<<"\
                   moves quasi-plateau"<<movesOnQuasiPlateau(iter,obj)<<":plat Heigh diff"<<quasiPlatHeighDiff;
            if(maxRadius!=INFTY_RADIUS_NO_DISTS){
                clog<<"|dstToCenter:"<<currDist;
                clog<<"|repulsiveForce:"<<repulsiveForce;
            }
            clog<<endl;;
        }

        //1. choose non tabu move: select out deselect in; 
        int out =-1, in=-1;
        double bestContrib = INT_MIN;
        int      bstRnd = 0;           //equal-quality moves ranked by taking max rand nr generated per move
        for(int i=0;i<(::n);i++)
            if((currSol[i])&&(tabu[i]<iter))
                for(int j=0;j<(::n);j++)
                    if((!currSol[j])&&(tabu[j]<iter)) 
                        updateBestMove(i,j);

        if(in<0) {               //everything tabu
            clog<<"all tabu\n";
            continue;
        }
        assert(in<(::n));

        //2. perform move
        currSol[in]  = 0;
        currSol[out] = 1;
        obj += bestContrib;

        //3. set tabu list lengths
        tabu[in]  = iter+10+rand()%5;
        tabu[in] += min(::n/2,                          //vals >n/2 can make 
                        movesOnQuasiPlateau(iter,obj)); //all moves tabu
    
        tabu[out] = iter;

        if(iter>2*::n+random()%(::n))                   //Star repulsing after doing a no of iters in [n,2]
        if(maxRadius!=INFTY_RADIUS_NO_DISTS) {          //DGLS only
            //perform repulsion using Tabu lists
            for(int ii=0;ii<n;ii++)
                if( (currSol[ii]&&!center[ii])||        //selected vtx not in center: fix longer
                    (!currSol[ii]&&center[ii])  )       //nonselected vtx of center : fix longer
                        tabu[ii]=max(tabu[ii],iter+(int)repulsiveForce);

            if( (obj==startObj) && (obj==bestObj))      //when reaching same obj as center & bst
            if(currDist<maxRadius/2)                    //if still too close to center 
               repulsiveForce+=1.0/(double)(currDist+3);//increase repForce proport to dist to center

            //if(repulsiveForce>1)                      //Use this if you want complete repulsion, to a solution that 
            //    maxRadius=2*::k;                      //shares no vertex with the center, like a restart?

        }

        //4. update contrib useful for streamlinging
        for(int i=0;i<(::n);i++) {
            contrib[i] += ww[i][out];
            contrib[i] -= ww[i][in];
        }

        //5. update current distance, if maxRadius!=INFTY_RADIUS_NO_DISTS (std Tabu search)
        if(maxRadius!=INFTY_RADIUS_NO_DISTS) {
            /*currDist+=deltaDist(out,in,center);
              see definition of deltaDist() in spheres.cpp to  
              see why it is equivalent to below formula    */
            currDist+=2*(center[in]-center[out]);
        }

        //6. update best solution ever seen
        //TODO: use a function for below if that decides when to update bst solution
        if( (obj>bestObj) ||
            ( (obj==bestObj) && (currDist>=bestDist) )                                   ||//std LS: 2nd cond always true
            ( (currDist>maxRadius*3/4) && (bestDist<maxRadius/2) && (bestObj==startObj) )||//std LS: 2nd cond fails
            ( (repulsiveForce>0) && (bestDist<=maxRadius*9/10))                            //when repulsion, take a far solution >9/10radius
          ){
                 if(obj>bestObj){
                     clog<<"LS iter "<<iter<<" with new best obj="<<obj<<"|    currDist="<<currDist
                         <<", move in/out"<<in<<"/"<<out<<endl;
                 }   
                 bestObj  = obj;
                 bestDist = currDist;
                 for(int i=0;i<(::n);i++)
                     solbest [i] = currSol[i];
        }

    }//end for over iterations 
    if(repulsiveForce>0) {
        clog<<"repulse:"<<repulsiveForce<<"iter="<<iter<<"|";
        clog.flush();
    }
    if(maxRadius!=INFTY_RADIUS_NO_DISTS)
        clog<<"bestDist="<<bestDist<<" : ";
    return bestObj;
}


/*--------------------------     ALLOC/INIT and FREE Mem     --------------------------*/

void allocGlobalsTabuSearch()
{
    if(::tabu!=NULL)
        return;
    ::currSol  = new int[::n];
    ::tabu       = new int[::n];
    ::contrib    = new double[::n];  
    ::ww         = new double*[::n];
    for(int i=0;i<(::n);i++)
        ::ww[i] = new double[::n];
    for(int i=0;i<(::n);i++){
        for(int j=0;j<(::n);j++){
            if(::edge[i][j])
                //::ww[i][j ]=(i+j)%5;         
                ::ww[i][j ]=1;         //classical one
                //use  ww[i][j ]=(w[i]+w[j]) if the edge weights come from vtx weights and divide it later by k-1, because a k-cluster adds w[i] k-1 times in the k-1 edges linking i
            else
                ::ww[i][j] = 0;
        }
        ww[i][i] = 0;
    }
}
void freeGlobalsTabuSearch()
{
    delete[] ::tabu;
    delete[] ::contrib;
    delete[] ::currSol;
    for(int i=0;i<(::n);i++)
        delete[] ::ww[i];
    ::tabu = NULL;
}


/*-------------------------------------------------------------------------------------------------+
|                                                                                                  |
|                                         Graph readers                                            |
|                                                                                                  |
+-------------------------------------------------------------------------------------------------*/
void readGraph(char*filename)
{
     ifstream graphStream;
     graphStream.open(filename);
     if ( graphStream.fail() ){
           cerr << " Error opening " << filename << endl;
           exit(0);
     }
     char token[100];
     graphStream >> token;
     if (graphStream.eof()){
           cerr << " Found EOF while reading in the beginning " << filename << endl;
           exit(0);
     }
     int x1, x2;     
     int mCount=0;
     while ( ! graphStream.eof() ) {
        if (!strcmp(token, "p" )) {
            char dummyWord[50];
            graphStream >> dummyWord >> ::n >> ::m;
            clog << "Vertices = " << ::n << ",";
            clog << "Edges = " << ::m << "\n";               
            ::edge=new int*[::n];
            for(int ii=0;ii<(::n);ii++){
                ::edge[ii]=new int[::n];
                for(int jj=0;jj<(::n);jj++)
                    ::edge[ii][jj]=0;
            }
        }                                  
        if (!strcmp(token, "e")){
            graphStream >> x1 >> x2;
            x1--; x2--;
            if ( x1<0 || x2<0 || x1>=n || x2 >=n ){
                  cerr << "Error reading ::edges x1=" << x1 << ", x2=" << x2 << endl;
            }
            ::edge[x1][x2]=1;
            ::edge[x2][x1]=1;
            mCount++;
        }
        graphStream >> token;
     }  
     if ( (::m != mCount))
           cerr<<"Warning reading graph. Announced m size=" << ::m << ", found=" << mCount  << endl;
     ::m = mCount;
     graphStream.close();   
};
